__title__ = 'sail'
__description__ = 'CLI tool to deploy and manage WordPress applications on DigitalOcean'
__version__ = '0.9.15'
__author__ = 'Konstantin Kovshenin'
__author_email__ = 'kovshenin@gmail.com'
__url__ = 'https://sailed.io'
__license__ = 'GPLv3'
__copyright__ = 'Copyright 2021 Konstantin Kovshenin'
